/*
 * Low-Level PCI and SB support for BCM47xx (Linux support code)
 *
 * Copyright 2003, Broadcom Corporation
 * All Rights Reserved.                
 *                                     
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Broadcom Corporation;   
 * the contents of this file may not be disclosed to third parties, copied
 * or duplicated in any form, in whole or in part, without the prior      
 * written permission of Broadcom Corporation.                            
 *
 * $Id: pcibios.c,v 1.6 2003/04/23 00:17:57 mhuang Exp $ 
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/paccess.h>

#include <asm/bcm963xx/6348_intr.h>
#include <asm/bcm963xx/6348_map_part.h>
#include <asm/bcm963xx/board.h>

#define BCM6348_A0          /* apply to A0 chip only */

#define pciMemRange1        PCI_SIZE_16M  /* 16 MB */
#define pciMemBase1         0x08000000
#define pciRemap1           0x08000000
#define pciMemRange2        PCI_SIZE_16M  /* 16 MB */
#define pciMemBase2         0x09000000
#define pciRemap2           0x09000000
#define pciIoRange          PCI_SIZE_64K  /* 64 KB */
#define pciIoBase           0x0C000000
#define pciIoRemap          0x0C000000

#define pciAddrMask         0x1FFFFFFF

#define sysMemSpace1        0x10000000
#define sysMemSpace2        0x00000000

/* 
 * EBI bus clock is 33MHz and share with PCI bus
 * each clock cycle is 30ns.
 */
/* attribute memory access wait cnt for 4306 */
#define PCMCIA_ATTR_CE_HOLD     3  // data hold time 70ns
#define PCMCIA_ATTR_CE_SETUP    3  // data setup time 50ns
#define PCMCIA_ATTR_INACTIVE    6  // time between read/write cycles 180ns. For the total cycle time 600ns (cnt1+cnt2+cnt3+cnt4)
#define PCMCIA_ATTR_ACTIVE      10 // OE/WE pulse width 300ns

/* common memory access wait cnt for 4306 */
#define PCMCIA_MEM_CE_HOLD      1  // data hold time 30ns
#define PCMCIA_MEM_CE_SETUP     1  // data setup time 30ns
#define PCMCIA_MEM_INACTIVE     2  // time between read/write cycles 40ns. For the total cycle time 250ns (cnt1+cnt2+cnt3+cnt4)
#define PCMCIA_MEM_ACTIVE       5  // OE/WE pulse width 150ns

#define PCCARD_VCC_MASK     0x00070000  // Mask Reset also
#define PCCARD_VCC_33V      0x00010000
#define PCCARD_VCC_50V      0x00020000

typedef enum {
    MPI_CARDTYPE_NONE,      // No Card in slot
    MPI_CARDTYPE_PCMCIA,    // 16-bit PCMCIA card in slot    
    MPI_CARDTYPE_CARDBUS,   // 32-bit CardBus card in slot
}   CardType;

typedef enum {
    MPI_CARDVCC_00,
    MPI_CARDVCC_33,
    MPI_CARDVCC_50,
}   CardVcc;

typedef struct {
    BOOL    CardDetected;   // Card's been detected already
    uint32  CardType;       // 0 = None, 1 = PCMCIA, 2 = CardBus
    uint32  CardVcc;        // Card's voltage setting
    uint32  CardState;      // 0 = Disabled, 1 = Enabled
}   Context;

#define CARDBUS_SLOT        0    // Slot 0 is default for CardBus
#define MAX_DEVICE_NUM      21   // Slot 0 reserve for CardBus

#define PCI_CFG_6348(d, f, o)  ( (d << 11) | (f << 8) | (o/4 << 2) )

#define pcmciaAttrOffset    0x00200000
#define pcmciaMemOffset     0x00000000
// Needs to be right above PCI I/O space. Give 0x8000 (32K) to PCMCIA. 
#define pcmciaIoOffset      (pciIoBase + 0x80000)
// Base Address is that mapped into the MPI ChipSelect registers. 
// UBUS bridge MemoryWindow 0 outputs a 0x00 for the base.
#define pcmciaBase          0xbf000000
#define pcmciaAttr          (pcmciaAttrOffset | pcmciaBase)
#define pcmciaMem           (pcmciaMemOffset  | pcmciaBase)
#define pcmciaIo            (pcmciaIoOffset   | pcmciaBase)

/* Global SB handle */
spinlock_t mpi_lock = SPIN_LOCK_UNLOCKED;
static volatile MpiRegisters * mpi = (MpiRegisters *)(MPI_BASE);

static void mpi_SetupPciConfigAccess(uint32 addr)
{
    mpi->l2pcfgctl = (DIR_CFG_SEL | DIR_CFG_USEREG | addr) & ~CONFIG_TYPE;
}

static void mpi_ClearPciConfigAccess(void)
{
    mpi->l2pcfgctl = 0x00000000;
}

static void mpi_SetLocalPciConfigReg(uint32 reg, uint32 value)
{
    /* write index then value */
    mpi->pcicfgcntrl = PCI_CFG_REG_WRITE_EN + reg;;
    mpi->pcicfgdata = value;
}

static uint32 mpi_GetLocalPciConfigReg(uint32 reg)
{
    /* write index then get value */
    mpi->pcicfgcntrl = PCI_CFG_REG_WRITE_EN + reg;;
    return mpi->pcicfgdata;
}

static int mpi_GetPciConfigReg(int bus, int dev, int func, int offset, void *data, int len)
{
    volatile unsigned char *ioBase = (unsigned char *)(mpi->l2piobase | 0xA0000000);
    uint8 *p8 = (uint8 *)data;
    uint16 *p16 = (uint16 *)data;
    uint32 *p32 = (uint32 *)data;
    int bitoffset;
    uint32 regval;

    memset(data, 0xff, len);

    if (bus)
        return -1;

    mpi_SetupPciConfigAccess(PCI_CFG_6348(dev, func, offset));
    regval = *(uint32 *)ioBase;
    bitoffset = (offset % 4) * 8;
    switch(len) {
        case 1:
            *p8 =  (regval >> bitoffset);
            break;
        case 2:
            *p16 = (regval >> bitoffset);
            break;
        case 4:
            *p32 = (regval >> bitoffset);
            break;
        default:
            break;
    }
    mpi_ClearPciConfigAccess();

    return 0;
}

static int mpi_SetPciConfigReg(int bus, int dev, int func, int offset, void *data, int len)
{
    volatile unsigned char *ioBase = (unsigned char *)(mpi->l2piobase | 0xA0000000);
    uint32 data8 = *(uint8 *)data;
    uint32 data16 = *(uint16 *)data;
    uint32 data32 = *(uint32 *)data;
    int bitoffset;
    uint32 regval;
    uint32 mask;

    if (bus)
        return -1;

    mpi_SetupPciConfigAccess(PCI_CFG_6348(dev, func, offset));
    regval = *(uint32 *)ioBase;
    bitoffset = (offset % 4) * 8;
    switch(len) {
        case 1:
            mask = 0x000000ff << bitoffset;
            regval &= ~mask;
            regval |= (data8 << bitoffset);
            *(uint32 *)ioBase = regval;
            break;
        case 2:
            mask = 0x0000ffff << bitoffset;
            regval &= ~mask;
            regval |= (data16 << bitoffset);
            *(uint32 *)ioBase = regval;
            break;
        case 4:
            *(uint32 *)ioBase = data32;
            break;
        default:
            break;
    }
    mpi_ClearPciConfigAccess();

    return 0;
}

static int
mpi_read_config_byte(struct pci_dev *dev, int where, u8 *value)
{
    unsigned long flags;
    int ret;

    spin_lock_irqsave(&mpi_lock, flags);
    ret = mpi_GetPciConfigReg(dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn), where, value, sizeof(*value));
    spin_unlock_irqrestore(&mpi_lock, flags);
    return ret ? PCIBIOS_DEVICE_NOT_FOUND : PCIBIOS_SUCCESSFUL;
}

static int
mpi_read_config_word(struct pci_dev *dev, int where, u16 *value)
{
    unsigned long flags;
    int ret;

    spin_lock_irqsave(&mpi_lock, flags);
    ret = mpi_GetPciConfigReg(dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn), where, value, sizeof(*value));
    spin_unlock_irqrestore(&mpi_lock, flags);
    return ret ? PCIBIOS_DEVICE_NOT_FOUND : PCIBIOS_SUCCESSFUL;
}

static int
mpi_read_config_dword(struct pci_dev *dev, int where, u32 *value)
{
    unsigned long flags;
    int ret;

    spin_lock_irqsave(&mpi_lock, flags);
    ret = mpi_GetPciConfigReg(dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn), where, value, sizeof(*value));
    spin_unlock_irqrestore(&mpi_lock, flags);
    return ret ? PCIBIOS_DEVICE_NOT_FOUND : PCIBIOS_SUCCESSFUL;
}

static int
mpi_write_config_byte(struct pci_dev *dev, int where, u8 value)
{
    unsigned long flags;
    int ret;

    spin_lock_irqsave(&mpi_lock, flags);
    ret = mpi_SetPciConfigReg(dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn), where, (void *)&value, sizeof(value));
    spin_unlock_irqrestore(&mpi_lock, flags);
    return ret ? PCIBIOS_DEVICE_NOT_FOUND : PCIBIOS_SUCCESSFUL;
}

static int
mpi_write_config_word(struct pci_dev *dev, int where, u16 value)
{
    unsigned long flags;
    int ret;

    spin_lock_irqsave(&mpi_lock, flags);
    ret = mpi_SetPciConfigReg(dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn), where, (void *)&value, sizeof(value));
    spin_unlock_irqrestore(&mpi_lock, flags);
    return ret ? PCIBIOS_DEVICE_NOT_FOUND : PCIBIOS_SUCCESSFUL;
}

static int
mpi_write_config_dword(struct pci_dev *dev, int where, u32 value)
{
    unsigned long flags;
    int ret;

    spin_lock_irqsave(&mpi_lock, flags);
    ret = mpi_SetPciConfigReg(dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn), where, (void *)&value, sizeof(value));
    spin_unlock_irqrestore(&mpi_lock, flags);
    return ret ? PCIBIOS_DEVICE_NOT_FOUND : PCIBIOS_SUCCESSFUL;
}

static struct pci_ops pcibios_ops = {
    mpi_read_config_byte,
    mpi_read_config_word,
    mpi_read_config_dword,
    mpi_write_config_byte,
    mpi_write_config_word,
    mpi_write_config_dword
};

static int mpi_init(void)
{
    unsigned long data;
    /*
     * Init the pci interface 
     */
    data = GPIO->GPIOMode; // GPIO mode register
    data |= GROUP2_PCI; // PCI internal arbiter
    GPIO->GPIOMode = data; // PCI internal arbiter

    /*
     * In the BCM6348 CardBus support is defaulted to Slot 0
     * because there is no external IDSEL for CardBus.  To disable
     * the CardBus and allow a standard PCI card in Slot 0 
     * set the cbus_idsel field to 0x1f.
    */
    /*
    uData = mpi->pcmcia_cntl1;
    uData |= CARDBUS_IDSEL;
    mpi->pcmcia_cntl1 = uData;
    */

    // UBUS to PCI address range
    // 16MBytes Memory Window 1. Mask determines which bits are decoded.
    mpi->l2pmrange1 = pciMemRange1;
    // UBUS to PCI Memory base address. This is akin to the ChipSelect base
    // register. 
    mpi->l2pmbase1 = pciMemBase1 & pciAddrMask;
    // UBUS to PCI Remap Address. Replaces the masked address bits in the
    // range register with this setting. 
    // Also, enable direct I/O and direct Memory accesses
    mpi->l2pmremap1 = (pciRemap1 | MEM_WINDOW_EN);

    // 16MBytes Memory Window 2
    mpi->l2pmrange2 = pciMemRange2;
    // UBUS to PCI Memory base address. 
    mpi->l2pmbase2 = pciMemBase2 & pciAddrMask;
    // UBUS to PCI Remap Address
    mpi->l2pmremap2 = (pciRemap2 | MEM_WINDOW_EN);

    // Setup PCI I/O Window range. Give 32K to PCI I/O
    mpi->l2piorange = pciIoRange;
    // UBUS to PCI I/O base address 
    mpi->l2piobase = pciIoBase & pciAddrMask;
    // UBUS to PCI I/O Window remap
    mpi->l2pioremap = (pciIoRemap | MEM_WINDOW_EN);

    // enable PCI related GPIO pins and data swap between system and PCI bus
    mpi->locbuscntrl = (EN_PCI_GPIO | DIR_U2P_NOSWAP);

    /* Enable 6348 BusMaster and Memory access mode */
    data = mpi_GetLocalPciConfigReg(PCI_COMMAND);
    data |= (PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER);
    mpi_SetLocalPciConfigReg(PCI_COMMAND, data);

    /* Configure two 16 MByte PCI to System memory regions. */
    /* These memory regions are used when PCI device is a bus master */
    /* Accesses to the SDRAM from PCI bus will be "byte swapped" for this region */
    mpi_SetLocalPciConfigReg(PCI_BASE_ADDRESS_3, sysMemSpace1);
    mpi->sp0remap = 0x0;

    /* Accesses to the SDRAM from PCI bus will not be "byte swapped" for this region */
    mpi_SetLocalPciConfigReg(PCI_BASE_ADDRESS_4, sysMemSpace2);
    mpi->sp1remap = 0x0;
    mpi->pcimodesel |= (PCI_BAR2_NOSWAP | 0x40);

    /*
     * Change 6348 PCI Cfg Reg. offset 0x40 to PCI memory read retry count infinity
     * by set 0 in bit 8~15.  This resolve read Bcm4306 srom return 0xffff in
     * first read.
     */
    data = mpi_GetLocalPciConfigReg(BRCM_PCI_CONFIG_TIMER);
    data &= ~BRCM_PCI_CONFIG_TIMER_RETRY_MASK;
    data |= 0x00000080;
    mpi_SetLocalPciConfigReg(BRCM_PCI_CONFIG_TIMER, data);

    /* enable pci interrupt */
    mpi->locintstat |= (EXT_PCI_INT << 16);

    return 0;
}

void __init
pcibios_init(void)
{
    ulong flags;

    spin_lock_init(&mpi_lock);

    spin_lock_irqsave(&mpi_lock, flags);
    mpi_init();
    spin_unlock_irqrestore(&mpi_lock, flags);

    /* Scan the SB bus */
    pci_scan_bus(0, &pcibios_ops, NULL);
}

char * __init
pcibios_setup(char *str)
{
    if (!strncmp(str, "ban=", 4)) {
        //mpi_ban(simple_strtoul(str + 4, NULL, 0));
        return NULL;
    }

    return (str);
}

static u32 pci_iobase = pciIoRemap;
static u32 pci_membase = pciRemap1;

void __init
pcibios_fixup_bus(struct pci_bus *b)
{
    struct list_head *ln;
    struct pci_dev *d;
    struct resource *res;
    int pos, size;
    u32 *base;
    char irq;

    printk("PCI: Fixing up bus %d\n", b->number);

    /* Fix up external PCI */
    for (ln=b->devices.next; ln != &b->devices; ln=ln->next) {
        d = pci_dev_b(ln);
        /* Fix up resource bases */
        for (pos = 0; pos < 4; pos++) {
            res = &d->resource[pos];
            base = (res->flags & IORESOURCE_IO) ? &pci_iobase : &pci_membase;
            if (res->end) {
                size = res->end - res->start + 1;
                if (*base & (size - 1))
                    *base = (*base + size) & ~(size - 1);
                res->start = *base;
                res->end = res->start + size - 1;
                *base += size;
                pci_write_config_dword(d, PCI_BASE_ADDRESS_0 + (pos << 2), res->start);
            }
            /* Fix up PCI bridge BAR0 only */
            if (b->number == 0 && PCI_SLOT(d->devfn) == 0)
                break;
        }
        /* Fix up interrupt lines */
        pci_write_config_byte(d, PCI_INTERRUPT_LINE, INTERRUPT_ID_MPI);
        pci_read_config_byte(d, PCI_INTERRUPT_LINE, &irq);
        d->irq = irq;
    }
}

unsigned int pcibios_assign_all_busses(void)
{
    return 1;
}

void pcibios_align_resource(void *data, struct resource *res,
			    unsigned long size)
{
}

int pcibios_enable_resources(struct pci_dev *dev)
{
    u16 cmd, old_cmd;
    int idx;
    struct resource *r;

    if (dev->bus->number != 0)
        return 0;

    pci_read_config_word(dev, PCI_COMMAND, &cmd);
    old_cmd = cmd;
    for(idx=0; idx<5; idx++) {
        r = &dev->resource[idx];
        if (r->flags & IORESOURCE_IO)
            cmd |= PCI_COMMAND_IO;
        if (r->flags & IORESOURCE_MEM)
            cmd |= PCI_COMMAND_MEMORY;
    }
    if (dev->resource[PCI_ROM_RESOURCE].start)
        cmd |= PCI_COMMAND_MEMORY;
    if (cmd != old_cmd) {
        printk("PCI: Enabling device %s (%04x -> %04x)\n", dev->slot_name, old_cmd, cmd);
        pci_write_config_word(dev, PCI_COMMAND, cmd);
    }
    return 0;
}

int pcibios_enable_device(struct pci_dev *dev)
{
    /* External PCI device enable */
    if (dev->bus->number == 0)
        return pcibios_enable_resources(dev);

    return 0;
}

void pcibios_update_resource(struct pci_dev *dev, struct resource *root,
			     struct resource *res, int resource)
{
    unsigned long where, size;
    u32 reg;

    /* External PCI only */
    if (dev->bus->number != 0)
        return;

    where = PCI_BASE_ADDRESS_0 + (resource * 4);
    size = res->end - res->start;
    pci_read_config_dword(dev, where, &reg);
    reg = (reg & size) | (((u32)(res->start - root->start)) & ~size);
    pci_write_config_dword(dev, where, reg);
}

static void __init quirk_mpi_bridge(struct pci_dev *dev)
{
    if (dev->bus->number != 0 || PCI_SLOT(dev->devfn) != 0)
        return;

    /* Enable PCI bridge bus mastering and memory space */
    pci_set_master(dev);
    pcibios_enable_resources(dev);
}	

struct pci_fixup pcibios_fixups[] = {
    { PCI_FIXUP_HEADER, PCI_ANY_ID, PCI_ANY_ID, quirk_mpi_bridge },
    { 0 }
};
